#include <iostream>
#include <sndfile.h>
#include <dirent.h>

#include "system.h"

#include <cmath>
#include <gtkmm.h>

using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::vector;

#include "peak.h"
#include "ayyi.h"
#include "aubio.h"
#include "aubio/aubio.h"
#include "window.h"

#define eprintf(A, ...) errprintf("aubio", __func__, A, ##__VA_ARGS__)
#define w(A, ...) warnprintf("aubio", __func__, A, ##__VA_ARGS__)
#define dbg(A, ...) dbgprintf("aubio", __func__, A, ##__VA_ARGS__)
#define PF printf("%s()...\n", __func__);
#define PF_DONE printf("%s(): done.\n", __func__);

void printFeatures(int, int, int, Vamp::Plugin::FeatureSet);
void transformInput(float *, size_t);
void fft(unsigned int, bool, double *, double *, double *, double *);
void printPluginPath();

extern AppWindow*   window;
extern Ayyi_client* ayyi;
extern Aubio*       aubio;
extern Wave*        wave;


Aubio::Aubio()
	: mix(false), threshold(0.7)
{
	string soname = "/usr/lib/vamp/vamp-aubio.so";
	string plugid = "aubioonset";

	try {
		window->peak = new Peak(wave);
	} catch (...) {
		eprintf("peakfile error.\n");
		throw std::exception();
	}

	libraryHandle = DLOPEN(soname, RTLD_LAZY);

	if (!libraryHandle) {
		eprintf("Failed to open plugin library %s: %s\n", soname.c_str(), DLERROR());
		throw std::exception();
	}

	cerr << "Opened plugin library " << soname << endl;

	VampGetPluginDescriptorFunction fn = (VampGetPluginDescriptorFunction) DLSYM(libraryHandle, "vampGetPluginDescriptor");
    
	if (!fn) {
		cerr << __func__ << ": No Vamp descriptor function in library " << soname << endl;
		DLCLOSE(libraryHandle);
		throw std::exception();
	}

	//cerr << argv[0] << ": Found plugin descriptor function" << endl;

	int index = 0;
	int plugnumber = -1;
	const VampPluginDescriptor *descriptor = 0;

	while ((descriptor = fn(VAMP_API_VERSION, index))) {

        Vamp::PluginHostAdapter plugin(descriptor, 48000);
        //cerr << argv[0] << ": Plugin " << (index+1) << " is \"" << plugin.getIdentifier() << "\"" << endl;

        if (plugin.getIdentifier() == plugid) plugnumber = index;
        
        ++index;
	}

	//cerr << argv[0] << ": Done\n" << endl;

	if (wave->path == "") {
        DLCLOSE(libraryHandle);
		throw std::exception();
    }

    if (plugnumber < 0) {
        if (plugid != "") {
            cerr << "ERROR: No such plugin as " << plugid << " in library" << endl;
            DLCLOSE(libraryHandle);
			throw std::exception();
        } else {
            plugnumber = 0;
        }
    }

    descriptor = fn(VAMP_API_VERSION, plugnumber);
    if (!descriptor) {
        DLCLOSE(libraryHandle);
		throw std::exception();
    }

	if (!wave->frames) {
        DLCLOSE(libraryHandle);
		throw std::exception();
    }

    /*Vamp::PluginHostAdapter* */plugin = new Vamp::PluginHostAdapter(descriptor, wave->samplerate);

    //cerr << "Running " << plugin->getIdentifier() << "..." << endl;

    block_size = plugin->getPreferredBlockSize();
    step_size = plugin->getPreferredStepSize();

    //cerr << "Preferred block size = " << block_size << ", step size = " << stepSize << endl;

    if (block_size == 0) block_size = 1024;

    bool rightBlockSize = true;

    if (plugin->getInputDomain() == Vamp::Plugin::FrequencyDomain) {

        int p = 1, b = block_size;
        while (b) {
            p <<= 1;
            b >>= 1;
        }
        if (p != block_size * 2) {
            cerr << "WARNING: Plugin requested non-power-of-two block size of " << block_size << ",\nwhich is not supported by this host.  ";
            block_size = p;
            cerr << "Rounding up to " << block_size << "." << endl;
            rightBlockSize = false;
        }
        if (step_size == 0) step_size = block_size / 2;

    } else {

        if (step_size == 0) step_size = block_size;
    }

    int channels = wave->channels;

    filebuf = new float[block_size * channels];
    plugbuf = new float*[channels];
    for (int c = 0; c < channels; ++c) plugbuf[c] = new float[block_size + 2];

    //cerr << "Using block size = " << block_size << ", step size = " << stepSize << endl;

    int min_ch = plugin->getMinChannelCount();
    int max_ch = plugin->getMaxChannelCount();
    //cerr << "Plugin accepts " << min_ch << " -> " << max_ch << " channel(s)" << endl;

    Vamp::Plugin::OutputList outputs = plugin->getOutputDescriptors();
    Vamp::Plugin::OutputDescriptor od;

    //--------------------------------------

	/*
		aubio detection types:

		aubio_onset_energy 	    energy based
		aubio_onset_specdiff 	spectral diff
		aubio_onset_hfc 	    high frequency content
		aubio_onset_complex 	complex domain
		aubio_onset_phase 	    phase fast
		aubio_onset_kl 	        Kullback Liebler.
		aubio_onset_mkl 	    modified Kullback Liebler
	*/
	onset_types = new std::map<std::string, int>;
	onset_types->insert(std::pair<std::string, int>("Energy",               aubio_onset_energy));
	onset_types->insert(std::pair<std::string, int>("Spectral Diff",        aubio_onset_specdiff));
	onset_types->insert(std::pair<std::string, int>("HF Content",           aubio_onset_hfc));
	onset_types->insert(std::pair<std::string, int>("Complex Domain",       aubio_onset_complex));
	onset_types->insert(std::pair<std::string, int>("Phase Fast",           aubio_onset_phase));
	onset_types->insert(std::pair<std::string, int>("Kullback Liebler",     aubio_onset_kl));
	onset_types->insert(std::pair<std::string, int>("Kullback Liebler Mod", aubio_onset_kl));

    Vamp::Plugin::ParameterList p_list = plugin->getParameterDescriptors();
    for(Vamp::Plugin::ParameterList::iterator i=p_list.begin(); i!=p_list.end(); i++){
        cout << " " << (*i).name << ": identifier=" << (*i).identifier << " default=" << (*i).defaultValue << " ("<< (*i).minValue << " - " << (*i).maxValue << ")" << endl;
    }
    plugin->setParameter("onsettype", aubio_onset_mkl);
    plugin->setParameter("peakpickthreshold", threshold);
    cout << "type=" << plugin->getParameter("onsettype") << endl;
    cout << "threshhold=" << plugin->getParameter("peakpickthreshold") << endl;

    //--------------------------------------

    int returnValue = 1;

    int output = 0;

    if (min_ch > channels || max_ch < channels) {
        if (min_ch == 1) {
            cerr << "WARNING: Sound file has " << channels << " channels, mixing down to 1" << endl;
            mix = true;
            channels = 1;
        } else {
            cerr << "ERROR: Sound file has " << channels << " channels, out of range for plugin" << endl;
            goto done;
        }
    }

    if (outputs.empty()) {
        cerr << "Plugin has no outputs!" << endl;
        goto done;
    }

    if (int(outputs.size()) <= output) {
        cerr << "Output " << output << " requested, but plugin has only " << outputs.size() << " output(s)" << endl;
        goto done;
    }        

    od = outputs[output];
    cerr << "Output is " << od.identifier << endl;

    if (!plugin->initialise(channels, step_size, block_size)) {
        cerr << "ERROR: Plugin initialise (channels = " << channels << ", step_size = " << step_size << ", block_size = " << block_size << ") failed." << endl;
        if (!rightBlockSize) cerr << "(Probably because I couldn't provide the plugin's preferred block size.)" << endl;
        goto done;
    }

	aubio = this;
	window->on_aubio_init();
done:
	{}
}


Aubio::~Aubio()
{
	delete plugin;
	DLCLOSE(libraryHandle);
}


void
Aubio::run()
{
	out.clear(); //clear previous results.

    plugin->setParameter("peakpickthreshold", threshold);
	dbg("threshold=%.2f", threshold);

	//can we do this more than once? looks like we can
    if (!plugin->initialise(wave->channels, step_size, block_size)) {
        cerr << "ERROR: Plugin initialise (channels = " << wave->channels << ", step_size = " << step_size << ", block_size = " << block_size << ") failed." << endl;
        //if (!rightBlockSize) cerr << "(Probably because I couldn't provide the plugin's preferred block size.)" << endl;
        return;
    }

    int output = 0;
    int channels = wave->channels;

    for (size_t i = 0; i < wave->frames; i += step_size) {

        int count;

        if (sf_seek(wave->sndfile, i, SEEK_SET) < 0) { cerr << "ERROR: sf_seek failed: " << sf_strerror(wave->sndfile) << endl; break; }
        
        if ((count = sf_readf_float(wave->sndfile, filebuf, block_size)) < 0) {
            cerr << "ERROR: sf_readf_float failed: " << sf_strerror(wave->sndfile) << endl;
            break;
        }

        for (int c = 0; c < channels; ++c) {
            for (int j = 0; j < block_size; ++j) {
                plugbuf[c][j] = 0.0f;
            }
        }

        for (int c = 0; c < channels; ++c) {
            int tc = c;
            if (mix) tc = 0;
            for (int j = 0; j < block_size && j < count; ++j) {
                plugbuf[tc][j] += filebuf[j * channels + c];
            }

            if (plugin->getInputDomain() == Vamp::Plugin::FrequencyDomain) {
                transformInput(plugbuf[tc], block_size);
            }
        }

        process_output(i, plugin->process(plugbuf, Vamp::RealTime::frame2RealTime(i, wave->samplerate)));
		/*
		Vamp::Plugin::FeatureSet features = plugin->process(plugbuf, Vamp::RealTime::frame2RealTime(i, wave->samplerate));
        printFeatures(i, wave->samplerate, output, features);

		std::vector<Vamp::Plugin::Feature> feature = features[0];

		results = new Vamp::Plugin::FeatureSet(features); //FIXME free old results
		*/
    }

    printFeatures(wave->frames, wave->samplerate, output, plugin->getRemainingFeatures());
}


void
Aubio::set_threshold(double _threshold)
{
	threshold = _threshold;
	//run();
}


//std::list<std::string>*
std::map<std::string, int>*
Aubio::get_onset_types()
{
	return onset_types;
}


void
printPluginPath()
{
    vector<string> path = Vamp::PluginHostAdapter::getPluginPath();
    for (size_t i = 0; i < path.size(); ++i) {
        cerr << path[i] << endl;
    }
}

#ifdef HAVE_OPENDIR

void
enumeratePlugins()
{
    cerr << endl << "Vamp plugin libraries found in search path:" << endl;
    vector<string> path = Vamp::PluginHostAdapter::getPluginPath();
    for (size_t i = 0; i < path.size(); ++i) {
        cerr << "\n" << path[i] << ":" << endl;
        DIR *d = opendir(path[i].c_str());
        if (!d) {
            perror("Failed to open directory");
            continue;
        }
        struct dirent *e = 0;
        while ((e = readdir(d))) {
//            cerr << "reading: " << e->d_name << endl;
            if (!(e->d_type & DT_REG)) {
//                cerr << e->d_name << ": not a regular file" << endl;
                continue;
            }
            int len = strlen(e->d_name);
            if (len < int(strlen(PLUGIN_SUFFIX) + 2) ||
                e->d_name[len - strlen(PLUGIN_SUFFIX) - 1] != '.' ||
                strcmp(e->d_name + len - strlen(PLUGIN_SUFFIX), PLUGIN_SUFFIX)) {
//                cerr << e->d_name << ": not a library file" << endl;
                continue;
            }
            char *fp = new char[path[i].length() + len + 3];
            sprintf(fp, "%s/%s", path[i].c_str(), e->d_name);
            void *handle = DLOPEN(string(fp), RTLD_LAZY);
            if (handle) {
                VampGetPluginDescriptorFunction fn = (VampGetPluginDescriptorFunction) DLSYM (handle, "vampGetPluginDescriptor");
                if (fn) {
                    cerr << "\n  " << e->d_name << ":" << endl;
                    int index = 0;
                    const VampPluginDescriptor *descriptor = 0;
                    while ((descriptor = fn(VAMP_API_VERSION, index))) {
                        Vamp::PluginHostAdapter plugin(descriptor, 48000);
                        char c = char('A' + index);
                        if (c > 'Z') c = char('a' + (index - 26));
                        cerr << "    [" << c << "] [v"
                             << plugin.getVampApiVersion() << "] "
                             << plugin.getName()
                             << ", \"" << plugin.getIdentifier() << "\""
                             << " [" << plugin.getMaker()
                             << "]" << endl;
                        if (plugin.getDescription() != "") {
                            cerr << "        - " << plugin.getDescription() << endl;
                        }
                        Vamp::Plugin::OutputList outputs = plugin.getOutputDescriptors();
                        if (outputs.size() > 1) {
                            for (size_t j = 0; j < outputs.size(); ++j) {
                                cerr << "         (" << j << ") "
                                     << outputs[j].name
                                     << ", \"" << outputs[j].identifier << "\""
                                     << endl;
                                if (outputs[j].description != "") {
                                    cerr << "             - " << outputs[j].description << endl;
                                }
                            }
                        }
                        ++index;
                    }
                } else {
//                    cerr << e->d_name << ": no Vamp descriptor function" << endl;
                }
                DLCLOSE(handle);
            } else {
                cerr << "\n" << e->d_name << ": unable to load library (" << DLERROR() << ")" << endl;
            }                
        }
        closedir(d);
    }
    cerr << endl;
}

#endif


void
Aubio::process_output(int frame, Vamp::Plugin::FeatureSet features)
{
	int output = 0;
	int sr = wave->samplerate;
	Vamp::RealTime rt;

	//dbg("size=%i", features[output].size());
    for (unsigned int i = 0; i < features[output].size(); ++i) {
		//std::vector<Vamp::Plugin::Feature, std::allocator<Vamp::Plugin::Feature> >& f = features[output];
		Vamp::Plugin::Feature feature = features[output][i];
        //Vamp::RealTime rt = Vamp::RealTime::frame2RealTime(frame, sr);
        if (feature.hasTimestamp) {
			out.push_back(feature.timestamp);
            rt = feature.timestamp;
        } else cout << "no timestamp i";
        //cout << rt.toString() << endl;
    }
}


void
printFeatures(int frame, int sr, int output, Vamp::Plugin::FeatureSet features)
{
	//dbg("size=%i", features[output].size());
    for (unsigned int i = 0; i < features[output].size(); ++i) {
		//std::vector<Vamp::Plugin::Feature, std::allocator<Vamp::Plugin::Feature> >& f = features[output];
		Vamp::Plugin::Feature feature = features[output][i];
		aubio->out.push_back(feature.timestamp);

        Vamp::RealTime rt = Vamp::RealTime::frame2RealTime(frame, sr);
        if (features[output][i].hasTimestamp) {
            rt = features[output][i].timestamp;
        } else cout << "no timestamp i";
		/*
        cout << rt.toString() << ":";
        for (unsigned int j = 0; j < features[output][i].values.size(); ++j) {
            cout << " " << features[output][i].values[j];
        }
		*/
        cout << endl;
    }
}

void
transformInput(float *buffer, size_t size)
{
    double *inbuf = new double[size * 2];
    double *outbuf = new double[size * 2];

    // Copy across with Hanning window
    for (size_t i = 0; i < size; ++i) {
        inbuf[i] = double(buffer[i]) * (0.50 - 0.50 * cos(2 * M_PI * i / size));
        inbuf[i + size] = 0.0;
    }
    
    for (size_t i = 0; i < size/2; ++i) {
        double temp = inbuf[i];
        inbuf[i] = inbuf[i + size/2];
        inbuf[i + size/2] = temp;
    }

    fft(size, false, inbuf, inbuf + size, outbuf, outbuf + size);

    for (size_t i = 0; i <= size/2; ++i) {
        buffer[i * 2] = outbuf[i];
        buffer[i * 2 + 1] = outbuf[i + size];
    }
    
    delete[] inbuf;
    delete[] outbuf;
}


void
fft(unsigned int n, bool inverse, double *ri, double *ii, double *ro, double *io)
{
    if (!ri || !ro || !io) return;

    unsigned int bits;
    unsigned int i, j, k, m;
    unsigned int blockSize, blockEnd;

    double tr, ti;

    if (n < 2) return;
    if (n & (n-1)) return;

    double angle = 2.0 * M_PI;
    if (inverse) angle = -angle;

    for (i = 0; ; ++i) {
        if (n & (1 << i)) {
            bits = i;
            break;
        }
    }

    static unsigned int tableSize = 0;
    static int *table = 0;

    if (tableSize != n) {

	delete[] table;

	table = new int[n];

	for (i = 0; i < n; ++i) {
	
	    m = i;

	    for (j = k = 0; j < bits; ++j) {
		k = (k << 1) | (m & 1);
		m >>= 1;
	    }

	    table[i] = k;
	}

	tableSize = n;
    }

    if (ii) {
	for (i = 0; i < n; ++i) {
	    ro[table[i]] = ri[i];
	    io[table[i]] = ii[i];
	}
    } else {
	for (i = 0; i < n; ++i) {
	    ro[table[i]] = ri[i];
	    io[table[i]] = 0.0;
	}
    }

    blockEnd = 1;

    for (blockSize = 2; blockSize <= n; blockSize <<= 1) {

	double delta = angle / (double)blockSize;
	double sm2 = -sin(-2 * delta);
	double sm1 = -sin(-delta);
	double cm2 = cos(-2 * delta);
	double cm1 = cos(-delta);
	double w = 2 * cm1;
	double ar[3], ai[3];

	for (i = 0; i < n; i += blockSize) {

	    ar[2] = cm2;
	    ar[1] = cm1;

	    ai[2] = sm2;
	    ai[1] = sm1;

	    for (j = i, m = 0; m < blockEnd; j++, m++) {

			ar[0] = w * ar[1] - ar[2];
			ar[2] = ar[1];
			ar[1] = ar[0];

			ai[0] = w * ai[1] - ai[2];
			ai[2] = ai[1];
			ai[1] = ai[0];

			k = j + blockEnd;
			tr = ar[0] * ro[k] - ai[0] * io[k];
			ti = ar[0] * io[k] + ai[0] * ro[k];

			ro[k] = ro[j] - tr;
			io[k] = io[j] - ti;

			ro[j] += tr;
			io[j] += ti;
		}
	}

	blockEnd = blockSize;
	}

	if (inverse) {

		double denom = (double)n;

		for (i = 0; i < n; i++) {
			ro[i] /= denom;
			io[i] /= denom;
		}
	}
}


gboolean
on_plugin_done(gpointer data)
{
	window->glScene.update_lines();
	return false;
}


