#define UTILS_CC
#include <gtkmm.h>
#include <libart_lgpl/libart.h>
#include "utils.h"
#include "ayyi_types.h"

#define dbg(A, ...) dbgprintf("utils", __func__, A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("utils", __func__, A, ##__VA_ARGS__)
#define err(A, ...) errprintf("utils", __func__, A, ##__VA_ARGS__)

extern int debug;

void
dbgprintf(const char* claas, const char* func, char *format, ...)
{
	printf("%s::%s(): ", claas, func);

	va_list argp;           //points to each unnamed arg in turn
	va_start(argp, format); //make ap (arg pointer) point to 1st unnamed arg
	vprintf(format, argp);
	va_end(argp);           //clean up
	printf("\n");
}


void
dbgprintf2(const char* claas, const char* func, int level, char *format, ...)
{
	va_list argp;           //points to each unnamed arg in turn
	va_start(argp, format); //make ap (arg pointer) point to 1st unnamed arg
	if (level <= debug) {
		printf("%s::%s(): ", claas, func);
		vprintf(format, argp);
		va_end(argp);           //clean up
		printf("\n");
	}
}


void
errprintf(const char* claas, const char* func, char *format, ...)
{
  //fn prints an error string, then passes arguments on to vprintf.

  static char err[32] = "\x1b[1;31merror:\x1b[0;39m";
  printf("%s %s::%s(): ", err, claas, func);

  va_list argp;           //points to each unnamed arg in turn
  va_start(argp, format); //make ap (arg pointer) point to 1st unnamed arg
  vprintf(format, argp);
  va_end(argp);           //clean up
}


void
warnprintf(const char* claas, const char* func, char *format, ...)
{
  //fn prints an error string, then passes arguments on to vprintf.

  static char warn[32] = "\x1b[1;33mwarning:\x1b[0;39m";
  printf("%s %s::%s(): ", warn, claas, func);

  va_list argp;           //points to each unnamed arg in turn
  va_start(argp, format); //make ap (arg pointer) point to 1st unnamed arg
  vprintf(format, argp);
  va_end(argp);           //clean up
}


void
filename_change_extension(char* path, const char* extn)
{
  ASSERT_POINTER(path, "path");

  char* dot = strrchr(path, '.');
  char* ext;
  if(dot){
    ext = dot+1;
    strcpy(ext, extn);
  }else{
    warn("no file extension found.\n", "");
  }
}


void
color_rgba_to_gdk(GdkColor* colour, uint32_t rgba)
{
  ASSERT_POINTER(colour, "colour");

  colour->red   = (rgba & 0xff000000) >> 16;
  colour->green = (rgba & 0x00ff0000) >> 8;
  colour->blue  = (rgba & 0x0000ff00);
  //if(debug) printf("%s(): 0x%x->%i,%i,%i\n", __func__, rgba, colour->red, colour->green, colour->blue); 
}


void
pixbuf_fill_rgb(GdkPixbuf* pixbuf, GdkColor* colour)
{
  //this appears to just write over the whole pixbuf.

  //what format does the pixbuf have to be? 24bit?

  //printf("pixbuf_draw_rect()\n");
  art_u8 *buffer = (art_u8*)gdk_pixbuf_get_pixels(pixbuf);
  int bufwidth  = gdk_pixbuf_get_width(pixbuf);
  int bufheight = gdk_pixbuf_get_height(pixbuf);

  //GdkColor c = get_style_bg_color(GTK_STATE_NORMAL);
  //GdkColor c = {0,0,0xffff,0x8000};

  //art_rgba_run_alpha(buffer, colour->red/256, colour->green/256, colour->blue/256, 0xff, bufwidth*bufheight);

  art_rgb_fill_run(buffer, colour->red/256, colour->green/256, colour->blue/256, bufwidth*bufheight);
}


void
pixbuf_draw_line(GdkPixbuf* pixbuf, struct _ArtDRect* pts, double line_width, GdkColor* colour)
{
  //FIXME this leaks - dont know why - replace with cairo functions instead of libart.

  //the pixbuf buffer can be either 24 or 32bits per pixel.

  //this fns draw non-antialiased lines, so strange effects can occur with non-integer pts.

  art_u8 *buffer = (art_u8*)gdk_pixbuf_get_pixels(pixbuf);
  int bufwidth  = gdk_pixbuf_get_width    (pixbuf);
  int bufheight = gdk_pixbuf_get_height   (pixbuf);
  int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
  //printf("pixbuf_draw_line(): pixbuf width=%i height=%i\n", bufwidth, bufheight);

  art_u32 color_fg = ((colour->red/256)<< 24) | ((colour->green/256)<<16) | ((colour->blue/256)<<8) | (0xff); // RRGGBBAA
  //printf("%x ", color_fg);

  //make vector:
  static ArtVpath* vec = NULL;
  if(!vec) vec = art_new(ArtVpath, 10);
  vec[0].code = ART_MOVETO;
  vec[0].x = pts->x0;
  vec[0].y = pts->y0;
  vec[1].code = ART_LINETO;
  vec[1].x = pts->x1;
  vec[1].y = pts->y1;
  vec[2].code = ART_END;

  ArtSVP* svp = art_svp_vpath_stroke(vec,
                             ART_PATH_STROKE_JOIN_ROUND,//ArtPathStrokeJoinType join,
                             ART_PATH_STROKE_CAP_BUTT,//ArtPathStrokeCapType cap,
                             line_width,//double line_width,
                             1.0, //??????? double miter_limit,
                             1.0);//double flatness
  //render to buffer:
  art_rgb_svp_alpha(svp, 0, 0,
                    bufwidth, bufheight,//width, height,
                    color_fg, buffer,
                    rowstride,//number of bytes in each row.
                    NULL);
  //art_free(vec);
  art_free(svp);
}


GList*
get_dirlist(const char* path)
{
    /*
    scan a directory and return a list of any subdirectoies. Not recursive.
	-the list, and each entry in it,  must be freed.
    */

	GList* dir_list = NULL;
    char filepath[256];
    G_CONST_RETURN gchar *file;
    GError *error = NULL;
    GDir *dir;
    if((dir = g_dir_open(path, 0, &error))){
        while((file = g_dir_read_name(dir))){
            if(file[0]=='.') continue;
            snprintf(filepath, 128, "%s/%s", path, file);

            if(g_file_test(filepath, G_FILE_TEST_IS_DIR)){
                dbg("found dir: %s", filepath);
				dir_list = g_list_append(dir_list, g_strdup(filepath));
            }
        }
        g_dir_close(dir);
    }else{
        warn("cannot open directory. %s\n", error->message);
        g_error_free(error);
        error = NULL;
    }
	return dir_list;
}


bool
audio_path_get_leaf(const char* path, char* leaf)
{
  //gives the leafname contained in the given path.
  //-if the path is a directory, leaf should be empty, but without checking we have no way of knowing...
  //-relative paths are assumed to be relative to song->audiodir.
  //-fn is potentially slow as it accesses the disk.

  if(!strlen(path)){ err("path empty\n", ""); return false; }

  //look for slashes and chop off anything before:
  char* pos;
  if((pos = g_strrstr(path, "/"))){
    pos += 1; //move to the rhs of the slash.
  }else pos = (char*)path;

  //make leaf contain the last segment:
  g_strlcpy(leaf, pos, 128);
  //DBG (3, "pos=%s leaf=%s", pos, leaf);

  //DBG (2, "path=%s --> %s", path, leaf);
  return TRUE;
}


uint64_t
beats2mu(double beats)
{
  //note beats is float.

  return (uint64_t)(beats * AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT); //this is the definition of a mu.
}


void
mu2pos(uint64_t mu, struct _song_pos *pos)
{
    //ASSERT_POINTER(pos, "pos");

    uint32_t beats = mu / (AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT );

    uint32_t remainder  = mu % (AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT );
    uint32_t subs = remainder / AYYI_MU_PER_SUB;
    remainder = remainder % AYYI_MU_PER_SUB;

    pos->beat = beats;
    pos->sub  = subs;
    pos->mu   = remainder;
}


