/*
  This file is part of the Ayyi Project.
  copyright (C) 2007 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
	The project contains code from the Vamp:

    Vamp

    An API for audio analysis and feature extraction plugins.

    Centre for Digital Music, Queen Mary, University of London.
    Copyright 2006 Chris Cannam.
    FFT code from Don Cross's public domain FFT implementation.
  
    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR
    ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
    CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    Except as contained in this notice, the names of the Centre for
    Digital Music; Queen Mary, University of London; and Chris Cannam
    shall not be used in advertising or otherwise to promote the sale,
    use or other dealings in this Software without prior written
    authorization.
*/

#include <getopt.h>
#include <iostream>
#include <sndfile.h>
#include <dirent.h>

#include "system.h"

#include <cmath>
#include <gtkmm.h>

#include "config.h"
#include "utils.h"
#include "ardour.h"

using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::vector;

#include "aubio.h"
#include "peak.h"
#include "window.h"
#include "app.h"

#define HOST_VERSION "1.0"

#define eprintf(A, ...) errprintf("app", __func__, A, ##__VA_ARGS__)
#define w(A, ...) warnprintf("app", __func__, A, ##__VA_ARGS__)
#define dbg(A, ...) dbgprintf("app", __func__, A, ##__VA_ARGS__)
#define PF printf("%s()...\n", __func__);
#define PF_DONE printf("%s(): done.\n", __func__);

gpointer plugin_thread(gpointer data);
gboolean on_plugin_done(gpointer data);

AppWindow*   window;
Ayyi_ardour_client* ayyi;
Aubio*       aubio;
Wave*        wave;
Ayyi_client::Region* region;

GAsyncQueue* msg_queue  = NULL;
int          debug      = 0;

#ifdef HAVE_OPENDIR
void enumeratePlugins();
#endif


int
print_help ()
{
	cout << "Usage: " << PACKAGE_STRING << "\n"
	     << "  -r, --region    region index     The region number to edit. (default 0)\n"
	     << "  -v, --version                    Show version information\n"
	     << "  -h, --help                       Print this message\n"
	     << "  -d, --debug     level            Output debug info to stdout\n"
		;
	return 1;

}


int
parse_opts (int argc, char* argv[], int* region_idx)
{
	const char* optstring = "hvr:d:";

	const struct option longopts[] = {
		{ "region", 1, 0, 'r' },
		{ "version", 0, 0, 'v' },
		{ "help", 0, 0, 'h' },
		{ "debug", 1, 0, 'd' },
		{ 0, 0, 0, 0 }
	};

	int option_index = 0;
	int c = 0;

	while (1) {
		c = getopt_long (argc, argv, optstring, longopts, &option_index);

		if (c == -1) {
			break;
		}

		switch (c) {
		case 0:
			break;

		case 'r':
			*region_idx = atoi(optarg);
			dbg("region_idx=%i", *region_idx);
			break;

		case 'v':
			cout << "version " << PACKAGE_VERSION << endl;
			exit (0);
			break;

		case 'h':
			print_help ();
			exit (0);
			break;
		case 'd':
			debug = atoi(optarg);
			break;

		default:
			return print_help();
		}
	}

	if (optind < argc) {
		//session_name = argv[optind++];
	}


	return 0;
}


int main(int argc, char **argv)
{
#ifdef DEBUG
	g_log_set_always_fatal((GLogLevelFlags)(
	                      G_LOG_FLAG_RECURSION |
	                      G_LOG_FLAG_FATAL |
	                      G_LOG_LEVEL_ERROR |
	                      G_LOG_LEVEL_CRITICAL
	                       | G_LOG_LEVEL_WARNING
	                      ));
#endif
	int region_idx = 0;
	parse_opts (argc, argv, &region_idx);

	g_type_init(); 
	Glib::thread_init();

	ayyi = new Ayyi_ardour_client;
	if(!ayyi->connect()) return 1;
	if(!ayyi->get_shm()) return 1;

	region = new Ayyi_client::Region(region_idx);

	Gtk::Main kit(argc, argv);
	Gtk::GL::init(argc, argv);

#if 0
	int major, minor;
	Gdk::GL::query_version(major, minor);
	std::cout << "OpenGL extension version - " << major << "." << minor << std::endl;
#endif

#if 0
	TODO: do these links give a c++ way of doing the async queue?
		http://efax-gtk.cvs.sourceforge.net/*checkout*/efax-gtk/efax-gtk/src/utils/notifier.h
		http://efax-gtk.cvs.sourceforge.net/*checkout*/efax-gtk/efax-gtk/src/utils/notifier.cpp

		also need to see:

		http://efax-gtk.cvs.sourceforge.net/*checkout*/efax-gtk/efax-gtk/src/utils/io_watch.h
		http://efax-gtk.cvs.sourceforge.net/*checkout*/efax-gtk/efax-gtk/src/utils/io_watch.cpp
		which connects a file descriptor to the Glib main loop and
		http://efax-gtk.cvs.sourceforge.net/*checkout*/efax-gtk/efax-gtk/src/utils/pipes.h
		http://efax-gtk.cvs.sourceforge.net/*checkout*/efax-gtk/efax-gtk/src/utils/pipes.cpp
		which wraps a unix anonymous pipe.
#endif

	GError* error = NULL;
	msg_queue = g_async_queue_new();
	if(!g_thread_create(plugin_thread, NULL, FALSE, &error)){
		eprintf("error creating worker thread: %s\n", error->message);
		g_error_free(error);
	}

	std::list<filesource_shared*> file_list = ayyi->files_get_list();
	Partlist                    region_list = ayyi->get_parts_list();
	printf("song loaded. files=%i regions=%i\n%s%s\n", file_list.size(), region_list.size(), go_rhs, ok);

	try { wave = new Wave(region_idx); } catch (...) { return 1; }
	window = new AppWindow;
	aubio  = new Aubio();
	kit.run(*window);

    return 0;
}


gpointer
plugin_thread(gpointer data)
{
	if(!msg_queue){ eprintf("no msg_queue!\n"); return NULL; }

	g_async_queue_ref(msg_queue);

	GList* work_list = NULL;
	gpointer message;
	while(1){

		//any new work ?
		while(g_async_queue_length(msg_queue)){
			message = g_async_queue_pop(msg_queue);
			//dbg("new message! %p", message);
			aubio->set_threshold((double)GPOINTER_TO_INT(message) / 100);

			//clear old work list:
			g_list_free(work_list);
			work_list = NULL;

			work_list = g_list_append(work_list, message);
		}

		while(work_list/*!=NULL*/){
			//printf("overview_thread(): list length: %i.\n", g_list_length(msg_list));

			gpointer work = g_list_first(work_list)->data;
			//eprintf("%s(): sample=%p filename=%s.\n", __func__, sample, sample->filename);
			aubio->run();
			g_idle_add(on_plugin_done, NULL);

			work_list = g_list_remove(work_list, work);
		}
		usleep(500000);
	}
}


void
App::on_shm_complete()
{
	//ayyi->set_audiodir();
}


