/*
  This file is part of the Ayyi project. http://ayyi.org
  copyright (C) 2007-2009 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <fcntl.h>
#include <sys/time.h>
#include <iostream>
#include <exception>
#include <gtkmm.h>
#include <gtkglmm.h>
#include <GL/gl.h>
#include <GL/glu.h>
using std::cout;
using std::cerr;
using std::endl;
using std::string;
#include "aubio.h"
#include "ayyi.h"
#include "peak.h"
#include "window.h"

#define dbg(A, ...) dbgprintf("peak", __func__, (char*)A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("peak", __func__, (char*)A, ##__VA_ARGS__)
#define err(A, ...) errprintf("peak", __func__, (char*)A, ##__VA_ARGS__)
#define PF printf("peak::%s()...\n", __func__);

#define PEAK_VALUES_PER_SAMPLE 2    //one positive and one negative.
#define PEAK_RATIO 256
#define HAS_ALPHA_FALSE 0
#define BITS_PER_PIXEL 8

extern Ayyi_client* ayyi;
extern AppWindow*   window;

static gboolean block_load(gpointer _block_num);

struct peak_sample{
    short positive; //even numbered bytes in the src peakfile are positive peaks
    short negative; //odd  numbered bytes in the src peakfile are negative peaks
};

Peak::Peak(Wave* wav)
{
	this->wave = wav;

	//make new textures:
	//note: each block is for a fixed number of *samples*. They are independant of zoom.
	this->blocks = g_new0(struct _gl_blocks, 1);

	int ch_num = 0;
	char fullpath[256];
	wave->get_full_path(fullpath);
	dbg("ch=%i filename=%s path=%s", ch_num, wave->path.c_str(), fullpath);
	if(!get_filename(fullpath, ch_num)) throw std::exception();

	set_buflen();

	if(!load_peakfile()) throw std::exception();

	int num_blocks = num_peaks / PEAKS_PER_BLOCK;
	if(num_peaks % PEAKS_PER_BLOCK) num_blocks++;
	dbg("num_blocks=%i", num_blocks);
	blocks->size         = num_blocks;
	blocks->texture_name = g_new0(unsigned, num_blocks);

	glGenTextures(num_blocks, blocks->texture_name);

	int i; for(i=0;i<num_blocks;i++){
		g_idle_add(block_load, GINT_TO_POINTER(i)); //TODO can we get the Class version of the callback to work?
	}

	for(i=0;i<num_blocks;i++){
		if(!glIsTexture(blocks->texture_name[i])){ warn("texture not loaded. i=%i\n", i); break; }
	}
#ifdef TEST_TEXTURE_PERF
	GLboolean resident[num_blocks];
	if(glAreTexturesResident(num_blocks, blocks->texture_name, resident)) dbg(0, "all textures are resident.");
	else {
		DBG (0, "not all textures are resident.");
		int i; for(i=0;i<num_blocks;i++) printf("%i ", resident[i]);
		printf("\n");
	}
#endif
}


Peak::~Peak()
{
	//delete all the textures used by this part.
	glDeleteTextures(blocks->size, blocks->texture_name);
}


gboolean
Peak::load(gpointer _block_num)
{
	PF;
}


static gboolean
block_load(gpointer _block_num)
{
	int block_num = GPOINTER_TO_INT(_block_num);

	GdkPixbuf* pixbuf = window->peak->block_to_pixbuf(block_num);
	window->glScene.load_texture(window->peak->blocks->texture_name[block_num], pixbuf);
	gdk_pixbuf_unref(pixbuf);

	window->glScene.redraw();
	return false;
}


void
Peak::set_buflen()
{
	sf_count_t audio_length = wave->frames;

    int len_words;
    struct stat statbuf;
    if(valid){
      stat(filename.c_str(), &statbuf);
      len_words = statbuf.st_size / byte_depth; //buffer len. = length of peakfile in samples x2.
      if(statbuf.st_size<1)  { warn("bad file size: %i (%s)\n", (int)statbuf.st_size, filename.c_str()); return; }
      if(len_words>100000000){ warn("bad file size: %i (%s)\n", (int)statbuf.st_size, filename.c_str()); return; }
    }
    else len_words = 100;

    buflen    = len_words;
    num_peaks = len_words / PEAK_VALUES_PER_SAMPLE;
    dbg("%s: filesize=%i byte_depth=%i num_peaks=%i audio_length=%i", filename.c_str(), (int)statbuf.st_size, byte_depth, num_peaks, audio_length);

	if(num_peaks * PEAK_RATIO > audio_length){
#if 0
		warn("peakfile is too big. Ignoring excess data...\n");
#endif
		num_peaks = audio_length / PEAK_RATIO;
		buflen = num_peaks * PEAK_VALUES_PER_SAMPLE;
	}
}


bool
Peak::load_peakfile()
{
	int ch_num = 0;
	//int   peaktype = PEAKTYPE_SEISMIX;

	/*
	if(!g_file_test(fullpath, G_FILE_TEST_EXISTS)){ 
		warn("audio file doesnt exist! (%s)\n", fullpath);
		valid = FALSE;
	} else valid = TRUE;
	*/

	//find the name of the peakfile:
	char peakfile[256];
	strncpy(peakfile, filename.c_str(), 255);

	valid = TRUE;

	int fp = 0;
	if(valid){
		fp = open(peakfile, O_RDONLY);
		if(!fp){ warn("file open failure.\n", ""); return 0; }

		uint32_t bytes = buflen * byte_depth;
		buf0 = (short*)malloc(bytes);
		dbg("peakfile=%s buflen=%i", peakfile, buflen);

		//read the whole peak file into memory:
		if(read(fp, buf0, bytes) != bytes) err("read error. couldnt read %u bytes from '%s'\n", bytes, peakfile);

		close(fp);

		//DBG (2, "done. %s: %isamples (%li beats / %.3f secs).", filename, pool_item->peak_buflen, samples2beats(pool_item->peak_buflen), samples2secs(pool_item->peak_buflen*PEAK_RATIO));
	}
	valid = TRUE;
	return true;
}


bool
Peak::get_filename(const char* fullpath, int ch_num)
{
	//dbg("fullpath=%s", fullpath);

	char leafname[256];	audio_path_get_leaf(fullpath, leafname);

	char ardour_peakpath[256];
	snprintf(ardour_peakpath, 256, "%s/%s", ayyi->peaksdir.c_str(), leafname);

	//for some reason, peakfiles now have a "%A" in them!
	char* dot; if((dot = g_strrstr(ardour_peakpath, "."))) sprintf(dot, "%%A.");

	filename_change_extension(ardour_peakpath, "peak");
	if(!g_file_test(ardour_peakpath, G_FILE_TEST_EXISTS)){ 
		warn("cannot find peak file! (%s)\n", ardour_peakpath);
		return false;
	}else{
		//strcpy(peakfile, ardour_peakpath);
		filename = ardour_peakpath;
		valid  = TRUE;
		byte_depth = 4;
		use_float  = TRUE;
	}
	return true;
}


void
Peak::draw()
{
	//draw the waveform onto the window.
	//-each part is drawn as several rectangles, one per waveform texture.
	//-currently we have textures for the whole audio file, even though we only display part of it.

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glColor3f(1.0, 1.0, 1.0); //TODO color shouldnt make any difference with Blend disabled?

	double hzoom = 1.0f;

	const int samples_per_block = PEAK_RATIO * PEAKS_PER_BLOCK;
	const int first_block_num   = wave->region_start / samples_per_block;
	const int remainder         = wave->region_start % samples_per_block;
	double first_block_fraction = remainder / samples_per_block;

	double top        = window->glScene.get_height();
	double bot        = 0;
	double tot_wid    = window->glScene.get_width() * hzoom;
	//double tot_blocks = num_peaks / PEAKS_PER_BLOCK;
	double tot_blocks = (wave->region_end - wave->region_start) / (PEAK_RATIO * PEAKS_PER_BLOCK);
	double block_wid  = tot_wid / tot_blocks;

	double block_wid_px = block_get_width_px(0);
	//dbg("block_wid=%.4f block_wid_px=%.4f", block_wid, block_wid_px);
	double start      = remainder ? -block_wid_px + (first_block_fraction * block_wid_px) : 0;
	dbg("tot_blocks=%.4f region_start=%i start=%.2f", tot_blocks, wave->region_start, start);

	double tex_pct = 1.0; //use the whole texture
	int i; for(i=first_block_num; i<blocks->size; i++){
		bool is_last = (i == blocks->size - 1);

		glBindTexture(GL_TEXTURE_2D, blocks->texture_name[i]);

		double x = start + i * block_wid_px;
		if (is_last){
			//block_wid_px = block_wid = length - block_wid * (blocks->size - 1); //last block is smaller
			block_wid_px = block_wid = block_get_width_px(i);
			tex_pct = blocks->last_used - 0.05; //note compensation for 1px rounding error
			if(tex_pct < 0.0) tex_pct = 0;
		}
		/*if(i==0)*/ //dbg("%02i of %02i: is_last=%i x=%6.2f wid=%6.2f tex_pct=%.2f", i+1, blocks->size, is_last, x, block_wid, tex_pct);
		glBegin( GL_QUADS );
		glTexCoord2d(0.0,     0.0); glVertex2d(x+0.0,          top);
		glTexCoord2d(tex_pct, 0.0); glVertex2d(x+block_wid_px, top);
		glTexCoord2d(tex_pct, 1.0); glVertex2d(x+block_wid_px, bot);
		glTexCoord2d(0.0,     1.0); glVertex2d(x+0.0,          bot);
		glEnd();
	}
	//gl_print((start)/PX_TO_GL-4, bot/PX_TO_GL + text_height/2, gpart->length * arrange->hzoom, "%s", gpart->label);

	//length calculation checks:
	/*
	double tot_px_check = block_get_width_px(0) * (blocks->size - 1) + block_get_width_px(blocks->size - 1);
	dbg("check: tot=%.1f=%.1f", tot_px_check, tot_wid);
	*/
}


inline int
next_p2(int a)
{
	// Get The First Power Of 2 >= The Int That We Pass It.

	int rval=1;
	while(rval<a) rval<<=1;
	return rval;
}


GdkPixbuf*
Peak::block_to_pixbuf(int blocknum)
{
	//copy part of the audio peakfile at 1:1 scale to a pixbuf suitable for use a GL texture.

	gboolean is_last = (blocknum == blocks->size - 1);

	uint32_t colour = 0x0000ffff;
	GdkColor fg_colour;
	color_rgba_to_gdk(&fg_colour, colour);

	uint32_t bg_colour = 0x000000ff;

	int px_start =  blocknum      * PEAKS_PER_BLOCK;
	int px_stop  = (blocknum + 1) * PEAKS_PER_BLOCK;

	int width  = PEAKS_PER_BLOCK;
	if (is_last){
		int remaining     = num_peaks % PEAKS_PER_BLOCK;
		width             = next_p2(remaining);
		blocks->last_used = (double)remaining / (double)width;
		px_stop           = px_start + remaining;
	}
	/*if(blocknum < 10)*/ //dbg("%i: width=%i is_last=%i", blocknum, width, is_last);
	int height = 1024;
	GdkPixbuf* pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, HAS_ALPHA_FALSE, BITS_PER_PIXEL, width, height);

	//paint background:
	GdkColor test_colour = {0, 0xffff, 0x9999, 0x0000};
	pixbuf_fill_rgb(pixbuf, &test_colour);

	render_to_pixbuf(pixbuf, 0, &px_start, &px_stop, &fg_colour, bg_colour);

	//if(blocknum==0) gdk_pixbuf_save(pixbuf, "tmp.png", "png", NULL, NULL);

	return pixbuf;
}


struct _line{
	guchar a[1024]; //alpha level for each pixel in the line.
};
static struct _line line[3];

static void
line_write(struct _line* line, int index, guchar val)
{
	//remove fn once debugging complete?
	if(index < 0 || index >= 1024){ err("y=%i\n", index); return; }
	line->a[index] = val;
}


static void
line_clear(struct _line* line)
{
	int i;
	for(i=0;i<1024;i++) line->a[i] = 0;
}


static void
sort_(short* dest, const short* src, int size)
{
	//sort j into ascending order
	int i, j=0, min, new_min, top=size, p=0;
	short n[4] = {0,0,0,0};
	//printf("   size=%i", size);

	for(i=0;i<size;i++) n[i] = src[i]; //copy the source array

		for(i=0;i<size;i++){
			min = n[0];
			p=0;
			for(j=1;j<top;j++){
				new_min = MIN(min, n[j]);
				if(new_min < min){
				min = new_min;
				p = j;
			}
		}
		//printf("  %i min=%i p=%i", i, min, p);
		//p is the index to the value we have used, and need to remove.

		dest[i] = min;

		int m; for(m=p;m<top;m++) n[m] = n[m+1]; //move remaining entries down.
		//printf(" top=%i", top);
		top--;
		n[top] = 0;
	}
	//printf("  %i %i %i %i --> %i %i %i %i\n", src[0], src[1], src[2], src[3], dest[0], dest[1], dest[2], dest[3]);
}


void
Peak::render_to_pixbuf(GdkPixbuf* pixbuf, uint32_t region_idx, int* start, int* end, GdkColor* fg_colour, uint32_t colour_bg)
{
	//  -horizontal scaling is 1:1
	//  -the given pixbuf is for a single block only.
	//Can be simplified. Eg, the antialiasing should be removed.

	/*
     renders a peakfile (loaded into the buffer given by pool_item->buf) onto the given pixbuf.
     -the pixbuf must already be cleared, with the correct background colour.

     @param start - if set, we start rendering at this value from the left of the pixbuf.
     @param end   - if set, num of pixels from left of pixbuf to stop rendering at.
                    Can be bigger than pixbuf width.

	 further optimisation:
	 -dont scan pixels above the peaks. Should we record the overall region/file peak level?
	*/

	ASSERT_POINTER(pixbuf, "pixbuf");

	int bg_red = (colour_bg & 0xff000000) >> 24;
	int bg_grn = (colour_bg & 0x00ff0000) >> 16;
	int bg_blu = (colour_bg & 0x0000ff00) >>  8;
	int fg_red = fg_colour->red   >> 8;
	int fg_grn = fg_colour->green >> 8;
	int fg_blu = fg_colour->blue  >> 8;

	struct timeval time_start, time_stop;
	gettimeofday(&time_start, NULL);

	uint32_t start_offset = 0;//core_region_get_start_offset(region_idx) / PEAK_RATIO;               //FIXME

	struct peak_sample sample;
	//short sample_positive;    //even numbered bytes in the src peakfile are positive peaks;
	//short sample_negative;    //odd  numbered bytes in the src peakfile are negative peaks;
	short min;                //negative peak value for each pixel.
	short max;                //positive peak value for each pixel.
	if(!valid) return;
	//if(!pool_item->source_id){ err("bad core source id: %i.\n", pool_item->source_id); return; }

	short* buf       =         buf0; //source buffer.
	float* buf_float = (float*)buf0;
	//ASSERT_POINTER(buf, "buf");

	int width        = gdk_pixbuf_get_width(pixbuf);
	int height       = gdk_pixbuf_get_height(pixbuf);
	guchar* pixels   = gdk_pixbuf_get_pixels(pixbuf);
	int rowstride    = gdk_pixbuf_get_rowstride(pixbuf);
	if (height>1024) err("part too tall. not enough memory allocated.", "");
	int vscale = (256*128*2) / height;

	//printf("%s(): width=%i height=%i pixels=%p\n", __func__, width, height, pixels);
	int px=0;                       //pixel count starting at the lhs of the Part.
	int j,y;
	int srcidx_start=0;             //index into the source buffer for each sample pt.
	int srcidx_stop =0;   

	double xmag = 1; //making xmag smaller increases the visual magnification.
	                 //-the bigger the mag, the less samples we need to skip.
	                 //-as we're only dealing with smaller peak files, we need to adjust by the RATIO.

	//printf("%s(): xmag=%.2f srcbuflen=%i\n", __func__, xmag, buflen);
	if(buflen > 1000000){ err("buflen too big\n", ""); return; }

	int px_start = start ? *start : 0;
	int px_stop  = *end;
	//dbg("start=%i end=%i buflen(samples)=%i", px_start, px_stop, buflen/PEAK_VALUES_PER_SAMPLE);
	if(buflen/PEAK_VALUES_PER_SAMPLE < px_stop){ err("buffer too short for display range. buflen=%i end=%i\n", buflen/PEAK_VALUES_PER_SAMPLE, px_stop); return; }

	line_clear(&line[0]);
	int line_index = 0;

	for(px=px_start;px<px_stop;px++){
		srcidx_start = ((int)( px   * xmag)) + start_offset;
		srcidx_stop  = ((int)((px+1)* xmag)) + start_offset;
		//printf("s%i ", srcidx_start);

		struct _line* previous_line = &line[(line_index  ) % 3];
		struct _line* current_line  = &line[(line_index+1) % 3];
		struct _line* next_line     = &line[(line_index+2) % 3];

		//arrays holding subpixel levels.
		short lmax[4] = {0,0,0,0};
		short lmin[4] = {0,0,0,0};
		short k   [4] = {0,0,0,0}; //sorted copy of l.

		int mid = height / 2;

		if(srcidx_stop<buflen/2){
			min = 0; max = 0;
			int sub_px = 0;
			for(j=srcidx_start;j<srcidx_stop;j++){ //iterate over all the source samples for this pixel.
				if(use_float){
					//printf("%.1f ", buf_float[2*j]);
					sample.positive = (short)(buf_float[2*j +1] * (1 << 15)); //ardour peak files have negative peak first.
					sample.negative = (short)(buf_float[2*j   ] * (1 << 15));
				}else{
					sample.positive = buf[2*j   ];
					sample.negative = buf[2*j +1];
				}
				if(sample.positive > max) max = sample.positive;
				if(sample.negative < min) min = sample.negative;

				if(sub_px<4){
					//FIXME these supixels are not evenly distributed when > 4 available.
					lmax[sub_px] = (height * sample.positive) / (256*128*2);
					lmin[sub_px] =-(height * sample.negative) / (256*128*2); //lmin contains positive values.
				}
				sub_px++;
			}
			if(!sub_px) continue;

			if(!line_index){
				//first line - we also grab the previous sample for antialiasing.
				if(px){
					j = srcidx_start - 1;
					if(use_float){
						sample.positive = (short)(buf_float[2*j +1] * (1 << 15)); //ardour peak files have negative peak first.
						sample.negative = (short)(buf_float[2*j   ] * (1 << 15));
					}else{
						sample.positive = buf[2*j   ];
						sample.negative = buf[2*j +1];
					}
					max = sample.positive / vscale;
					min =-sample.negative / vscale;
					//printf(" j=%i max=%i min=%i\n", j, max, min);
					for(y=mid;y<mid+max;y++) line_write(previous_line, y, 0xff);
					for(y=mid;y<mid+max;y++) line_write(current_line, y, 0xff);
					for(y=mid;y>mid-max;y--) line_write(previous_line, y, 0xff);
					for(y=mid;y>mid-max;y--) line_write(current_line, y, 0xff);
				}
				else line_clear(current_line);
			}

			//scale the values to the part height:
			min = (height * min) / (256*128*2);
			max = (height * max) / (256*128*2);

			sort_(k, lmax, MIN(sub_px, 4));

			//note that we write into next_line, not current_line. There is a visible delay of one line.
			int alpha = 0xff;
			int v=0;
			int s,a;
			for(s=0;s<MIN(sub_px, 4);s++){
				//printf(" v=%i->%i ", v, k[s]);
				for(y=v;y<k[s];y++){
					line_write(next_line, mid +y, alpha);
				}
				v=k[s];
				alpha = (alpha * 2) / 3;
			}
			line_write(next_line, mid+k[s-1], alpha/2); //blur!
			for(y=k[s-1]+1;y<mid;y++){
				next_line->a[mid + y] = 0;
			}

			//negative peak:
			sort_(k, lmin, MIN(sub_px, 4));
			alpha = 0xff;
			v=mid;
			for(s=0;s<MIN(sub_px, 4);s++){
				//for(y=v;y>mid-k[s];y--) next_line->a[y] = alpha;
				for(y=v;y>mid-k[s];y--) line_write(next_line, y, alpha);
				v=mid-k[s];
				alpha = (alpha * 2) / 3;
			}
			line_write(next_line, mid-k[s-1], alpha/2); //antialias the end of the line.
			for(y=mid-k[s-1]-1;y>0;y--) line_write(next_line, y, 0);
			//printf("%.1f %i ", xf, height/2+min);

			//draw the lines:
			int blur = 6; //bigger value gives less blurring.
			for(y=0;y<height;y++){
				//if(y==0) printf("%i ", px-px_start);
				int p = 3*(px-px_start) + (height - y -1)*rowstride;
				if(p > rowstride*height || p < 0){ err("pixel index error! %i > %i px=%i(%i) y=%i row=%i width=%i rowstride=%i\n", p, 3*width*height, px, px-px_start, y, height-y-1, width, rowstride); return; }

				a = MIN((current_line->a[y] * 2)/3 + previous_line->a[y]/blur + next_line->a[y]/blur, 0xff);
				pixels[p  ] = (int)( bg_red * (0xff - a) + fg_red * a) >> 8;
				pixels[p+1] = (int)( bg_grn * (0xff - a) + fg_grn * a) >> 8;
				pixels[p+2] = (int)( bg_blu * (0xff - a) + fg_blu * a) >> 8;
			}

		}else{
			//no more source data available - as the pixmap is clear, we have nothing much to do.
			//gdk_draw_line(GDK_DRAWABLE(pixmap), gc, px, 0, px, height);//x1, y1, x2, y2
			printf("e ");
			struct _ArtDRect pts = {px, 0, px, height};
			pixbuf_draw_line(pixbuf, &pts, 1.0, fg_colour);
		}
		//next = srcidx + 1;
		//xf += PEAK_RATIO * PEAK_VALUES_PER_SAMPLE / samples_per_px;
		//printf("%.1f ", xf);

		line_index++;
		//printf("line_index=%i %i %i %i\n", line_index, (line_index  ) % 3, (line_index+1) % 3, (line_index+2) % 3);
	}

	//dbg("done. drawn: %i of %i stop=%i", line_index, width, srcidx_stop);

	gettimeofday(&time_stop, NULL);
	time_t secs = time_stop.tv_sec - time_start.tv_sec;
	suseconds_t usec;
	if(time_stop.tv_usec > time_start.tv_usec){
		usec = time_stop.tv_usec - time_start.tv_usec;
	}else{
		secs -= 1;
		usec = time_stop.tv_usec + 1000000 - time_start.tv_usec;
	}
	//printf("%s(): start=%03i: %lu:%06lu\n", __func__, px_start, secs, usec);
}


double
Peak::block_get_width_px(int blocknum)
{
	double hzoom        = 1.0f;
	double tot_wid      = window->glScene.get_width() * hzoom;

	if(blocks->size == 1) return tot_wid;

	if(	bool is_last = (blocknum == blocks->size - 1)){
		//subtract amount used so far from total
		return tot_wid - block_get_width_px(0) * (blocks->size - 1);
	}

	double peaks_per_px = num_peaks / tot_wid;

	return PEAKS_PER_BLOCK / peaks_per_px;
}


