#include "config.h"
#include <iostream>
#include <cstdlib>
#include <gtkmm.h>
#include <gtkglmm.h>

#include <GL/gl.h>
#include <GL/glu.h>

using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::vector;

#include "utils.h"
#include "aubio.h"
#include "peak.h"
#include "ayyi.h"
#include "window.h"

#define dbg(LEVEL, A, ...) dbgprintf2("window", __func__, LEVEL, (char*)A, ##__VA_ARGS__)
#define warn(A, ...) warnprintf("window", __func__, (char*)A, ##__VA_ARGS__)
#define err(A, ...) errprintf("window", __func__, (char*)A, ##__VA_ARGS__)
#define PF printf("window::%s()...\n", __func__);

extern Ayyi_client* ayyi;
extern AppWindow*   window;
extern Aubio*       aubio;
extern Wave*        wave;
extern Ayyi_client::Region* region;
extern GAsyncQueue* msg_queue;

///////////////////////////////////////////////////////////////////////////////
//
// OpenGL frame buffer configuration utilities.
//
///////////////////////////////////////////////////////////////////////////////

struct GLConfigUtil
{
  static void print_gl_attrib(const Glib::RefPtr<const Gdk::GL::Config>& glconfig,
                              const char* attrib_str,
                              int attrib,
                              bool is_boolean);

  static void examine_gl_attrib(const Glib::RefPtr<const Gdk::GL::Config>& glconfig);
};


// Print a configuration attribute.
void
GLConfigUtil::print_gl_attrib(const Glib::RefPtr<const Gdk::GL::Config>& glconfig,
                              const char* attrib_str,
                              int attrib,
                              bool is_boolean)
{
  int value;

  if (glconfig->get_attrib(attrib, value))
    {
      std::cout << attrib_str << " = ";
      if (is_boolean)
        std::cout << (value == true ? "true" : "false") << std::endl;
      else
        std::cout << value << std::endl;
    }
  else
    {
      std::cout << "*** Cannot get "
                << attrib_str
                << " attribute value\n";
    }
}

//
// Print configuration attributes.
//
void
GLConfigUtil::examine_gl_attrib(const Glib::RefPtr<const Gdk::GL::Config>& glconfig)
{
  std::cout << "\nOpenGL visual configurations :\n\n";

  std::cout << "glconfig->is_rgba() = "
            << (glconfig->is_rgba() ? "true" : "false")
            << std::endl;
  std::cout << "glconfig->is_double_buffered() = "
            << (glconfig->is_double_buffered() ? "true" : "false")
            << std::endl;
  std::cout << "glconfig->is_stereo() = "
            << (glconfig->is_stereo() ? "true" : "false")
            << std::endl;
  std::cout << "glconfig->has_alpha() = "
            << (glconfig->has_alpha() ? "true" : "false")
            << std::endl;
  std::cout << "glconfig->has_depth_buffer() = "
            << (glconfig->has_depth_buffer() ? "true" : "false")
            << std::endl;
  std::cout << "glconfig->has_stencil_buffer() = "
            << (glconfig->has_stencil_buffer() ? "true" : "false")
            << std::endl;
  std::cout << "glconfig->has_accum_buffer() = "
            << (glconfig->has_accum_buffer() ? "true" : "false")
            << std::endl;

  std::cout << std::endl;

  print_gl_attrib(glconfig, "Gdk::GL::USE_GL",           Gdk::GL::USE_GL,           true);
  print_gl_attrib(glconfig, "Gdk::GL::BUFFER_SIZE",      Gdk::GL::BUFFER_SIZE,      false);
  print_gl_attrib(glconfig, "Gdk::GL::LEVEL",            Gdk::GL::LEVEL,            false);
  print_gl_attrib(glconfig, "Gdk::GL::RGBA",             Gdk::GL::RGBA,             true);
  print_gl_attrib(glconfig, "Gdk::GL::DOUBLEBUFFER",     Gdk::GL::DOUBLEBUFFER,     true);
  print_gl_attrib(glconfig, "Gdk::GL::STEREO",           Gdk::GL::STEREO,           true);
  print_gl_attrib(glconfig, "Gdk::GL::AUX_BUFFERS",      Gdk::GL::AUX_BUFFERS,      false);
  print_gl_attrib(glconfig, "Gdk::GL::RED_SIZE",         Gdk::GL::RED_SIZE,         false);
  print_gl_attrib(glconfig, "Gdk::GL::GREEN_SIZE",       Gdk::GL::GREEN_SIZE,       false);
  print_gl_attrib(glconfig, "Gdk::GL::BLUE_SIZE",        Gdk::GL::BLUE_SIZE,        false);
  print_gl_attrib(glconfig, "Gdk::GL::ALPHA_SIZE",       Gdk::GL::ALPHA_SIZE,       false);
  print_gl_attrib(glconfig, "Gdk::GL::DEPTH_SIZE",       Gdk::GL::DEPTH_SIZE,       false);
  print_gl_attrib(glconfig, "Gdk::GL::STENCIL_SIZE",     Gdk::GL::STENCIL_SIZE,     false);
  print_gl_attrib(glconfig, "Gdk::GL::ACCUM_RED_SIZE",   Gdk::GL::ACCUM_RED_SIZE,   false);
  print_gl_attrib(glconfig, "Gdk::GL::ACCUM_GREEN_SIZE", Gdk::GL::ACCUM_GREEN_SIZE, false);
  print_gl_attrib(glconfig, "Gdk::GL::ACCUM_BLUE_SIZE",  Gdk::GL::ACCUM_BLUE_SIZE,  false);
  print_gl_attrib(glconfig, "Gdk::GL::ACCUM_ALPHA_SIZE", Gdk::GL::ACCUM_ALPHA_SIZE, false);

  std::cout << std::endl;
}


///////////////////////////////////////////////////////////////////////////////
//
// OpenGL drawing area.
//
///////////////////////////////////////////////////////////////////////////////

const Glib::ustring GLScene::m_FontString = "sans bold 9";


GLScene::GLScene()
  : m_FontListBase(0), m_FontHeight(0)
{
	// configure OpenGL-capable visual.
	Glib::RefPtr<Gdk::GL::Config> glconfig;

	// try double-buffered visual
	glconfig = Gdk::GL::Config::create(Gdk::GL::MODE_RGB    |
	                                   Gdk::GL::MODE_DEPTH  |
	                                   Gdk::GL::MODE_DOUBLE);
	if (!glconfig) {
		std::cerr << "*** Cannot find the double-buffered visual.\n"
		          << "*** Trying single-buffered visual.\n";

		// Try single-buffered visual
		if (!(glconfig = Gdk::GL::Config::create(Gdk::GL::MODE_RGB | Gdk::GL::MODE_DEPTH))){
			std::cerr << "*** Cannot find any OpenGL-capable visual.\n";
			std::exit(1);
		}
	}

	// print frame buffer attributes.
#ifdef DEBUGGG
	GLConfigUtil::examine_gl_attrib(glconfig);
#endif

	set_gl_capability(glconfig);

	//this->signal_client_event().connect(sigc::mem_fun(*this, &GLScene::on_mouse_event));

	add_events(Gdk::ALL_EVENTS_MASK);
}


GLScene::~GLScene()
{
}


void
GLScene::on_realize()
{
	// call the base on_realize()
	Gtk::DrawingArea::on_realize();

	// Get GL::Window.
	glwindow = get_gl_window();

	// *** OpenGL BEGIN ***
	if (!glwindow->gl_begin(get_gl_context())){ err("cant get gl_context\n"); return; }

	// Generate font display lists.
	m_FontListBase = glGenLists (128);

	Pango::FontDescription font_desc(m_FontString);

	Glib::RefPtr<Pango::Font> font = Gdk::GL::Font::use_pango_font(font_desc, 0, 128, m_FontListBase);
	if (!font) {
		std::cerr << "*** Can't load font "
		          << m_FontString
		          << std::endl;
		Gtk::Main::quit();
    }

	Pango::FontMetrics font_metrics = font->get_metrics();

	m_FontHeight = font_metrics.get_ascent() + font_metrics.get_descent();
	m_FontHeight = PANGO_PIXELS(m_FontHeight);

	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClearDepth(1.0);

	glViewport(0, 0, get_width(), get_height());

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, get_width(),
	        0.0, get_height(),
	       -1.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glwindow->gl_end();
	// *** OpenGL END ***

	set_size_request(20, -1);

	//test:
	//int w = get_width();
	//dbg(0, "w=%i", w);
}


bool
GLScene::on_configure_event(GdkEventConfigure* event)
{
  // Get GL::Window.
  Glib::RefPtr<Gdk::GL::Window> glwindow = get_gl_window();

  // *** OpenGL BEGIN ***
  if (!glwindow->gl_begin(get_gl_context())) return false;

  glViewport(0, 0, get_width(), get_height());

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, get_width(),
          0.0, get_height(),
          -1.0, 1.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glwindow->gl_end();
  // *** OpenGL END ***

  return true;
}


bool
GLScene::on_expose_event(GdkEventExpose* event)
{
	Glib::RefPtr<Gdk::GL::Window> glwindow = get_gl_window();

	if (!glwindow->gl_begin(get_gl_context())) return false;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/*
	// Draw some text.
	glColor3f(0.0, 0.0, 0.0);
	for (int i = 2; i >= -2; --i){
		glRasterPos2f(10.0, 0.5*get_height() + i * m_FontHeight);

		for (int j = ' '; j <= 'Z'; ++j) glCallList(m_FontListBase + j);
	}

	// Show font description string.
	glColor3f(1.0, 0.0, 0.0);
	glRasterPos2f(10.0, 10.0);
	glListBase(m_FontListBase);
	glCallLists(m_FontString.length(), GL_UNSIGNED_BYTE, m_FontString.c_str());
	*/

	window->peak->draw();
	print(5, 5, region->get_name());
	dbg(1, "region_start_pos=%u", region->get_pos());
	glDisable(GL_TEXTURE_2D); //test!

	dbg(2, "results map size: %i", window->result_cache.size());
	//if(window->current_lineset!=0 && window->current_lineset != window->lines.end()){
	if(window->result_cache.size()){
		std::list<Vamp::RealTime> list = window->current_lineset->second;
		/*
		for(std::list<Vamp::RealTime>::iterator i=list.begin(); i!=list.end(); i++){
			Vamp::RealTime rt = *i;
			printf("  %f %s\n", rt.msec(), rt.toString().c_str());
		}
		*/

		double max_x = 0.0f;
		//std::map<int, std::list<Vamp::RealTime> >::iterator first = window->lines.begin();
		//std::list<Vamp::RealTime> list = first->second;
		for(std::list<Vamp::RealTime>::iterator i=list.begin(); i!=list.end(); i++){
			Vamp::RealTime rt = *i;
			unsigned line_x_samples = rt.realTime2Frame(rt, wave->samplerate);
			//cout << " " << rt.realTime2Frame(rt, wave->samplerate);

			/*
			double tot_width = get_width(); //this will change when zoom is added.
			double x_scale = tot_width / wave->frames;
			double x = line_x_samples * x_scale;
			*/
			double x = line_x_samples / get_frames_per_px();
			if(x > max_x) max_x = x;
			glLineWidth( 1 );
			glColor3f(1.0, 0.0, 0.0);
			glBegin(GL_LINES);
			glVertex3f(x, 0, 0);
			glVertex3f(x, get_height(), 0);
			glEnd();
		}

		//check x calculations:
		if(max_x > get_width()) err("line x value out of range\n");
		double x_end = wave->frames / get_frames_per_px();
		dbg(3, "check: end=%.4f=%i", x_end, get_width());
	}

	rulerbar.draw();

	// Swap buffers.
	if (glwindow->is_double_buffered()) glwindow->swap_buffers();
	else glFlush();

	glwindow->gl_end();

	return true;
}


bool
GLScene::on_button_press_event(GdkEventButton* event)
{
	dbg(0, "on_button_press...");
	Gtk::DrawingArea::on_button_press_event(event);

	return false;
}


bool
GLScene::on_client_event(GdkEventClient* event)
{
	dbg(0, "on_client_event... never get here?");
	return false;
}


#define BUFSIZE 512
GLuint select_buf[BUFSIZE];

void
GLScene::start_picking(int cursor_x, int cursor_y)
{
	// from: http://www.lighthouse3d.com/opengl/picking/index.php?openglway2
	GLint viewport[4];

	glSelectBuffer(BUFSIZE, select_buf);
	glRenderMode(GL_SELECT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glGetIntegerv(GL_VIEWPORT, viewport);
	gluPickMatrix(cursor_x, viewport[3] - cursor_y, 5, 5, viewport);
	const int aspect_ratio = 1;
	gluPerspective(45, aspect_ratio, 0.1, 1000);
	glMatrixMode(GL_MODELVIEW);
	glInitNames();
}


void
GLScene::load_texture(int texture_name, GdkPixbuf* pixbuf)
{
	//load the pixbuf into the gl texture identified by texture_name.

	ASSERT_POINTER(pixbuf, "pixbuf");
	int width  = gdk_pixbuf_get_width (pixbuf);
	int height = gdk_pixbuf_get_height(pixbuf);

	//copy the image data to gl (can now be free'd):
	unsigned char* data = gdk_pixbuf_get_pixels(pixbuf);
	//if(texture_name<10) dbg("copying texture... data=%p width=%i texture_name=%u", data, width, texture_name);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	//gluBuild2DMipmaps( GL_TEXTURE_2D, 3, width, height, GL_RGB, GL_UNSIGNED_BYTE, data );

	//int bytes_per_px = 4;
	int bytes_per_px = 3;
	//int pixel_format = GL_RGBA; //GL_RGB
	int pixel_format = GL_RGB;
	glBindTexture(GL_TEXTURE_2D, texture_name);
	glTexImage2D(GL_TEXTURE_2D, 0, bytes_per_px, width, height, 0, pixel_format, GL_UNSIGNED_BYTE, data);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	//glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	//arg2? GLint components
	if(gluBuild2DMipmaps(GL_TEXTURE_2D, bytes_per_px, width, height, pixel_format, GL_UNSIGNED_BYTE, data)) warn("mipmap create error.\n", "");
}


void
GLScene::update_lines()
{
	//to be called following new data from plugin.
	dbg(2, "size=%i", aubio->out.size());

	int key = (int)(aubio->threshold * 100);
	std::map<const int, std::list<Vamp::RealTime> >::iterator existing = window->result_cache.find(key);
	if(existing == window->result_cache.end()){
		std::pair<std::_Rb_tree_iterator<std::pair<const int, std::list<Vamp::RealTime, std::allocator<Vamp::RealTime> > > >, bool> o = window->result_cache.insert(std::map<int, std::list<Vamp::RealTime> >::value_type(key, aubio->out));
		if(!o.second) err("insert failed?");
		else window->current_lineset = o.first;
	}
	else{
		dbg(0, "key=%i already have that one. Not adding...", key);
		window->current_lineset = existing;
	}

#if 0
	window->print_map();
#endif
	on_expose_event(NULL);

	/*
	dbg("checking...");
		std::map<int, std::list<Vamp::RealTime> >::iterator first = window->lines.begin();
		std::list<Vamp::RealTime> list = first->second;
		for(std::list<Vamp::RealTime>::iterator i=list.begin(); i!=list.end(); i++){
			Vamp::RealTime rt = *i;
			//printf("  %f %s\n", rt.msec(), rt.toString().c_str());
			//long realTime2Frame(const RealTime &r, unsigned int sampleRate);
			cout << "  " << rt.realTime2Frame(rt, wave->samplerate) << " " << rt.toString() << endl;
		}
	dbg("check done.");
	*/

	/*
	Vamp::Plugin::FeatureSet* results = aubio->results;
    int output = 0;
	dbg("size=%i", results[output].size());

	//std::vector<Vamp::Plugin::Feature>::iterator ifeature = results->begin();
	std::_Rb_tree_iterator<std::pair<const int, std::vector<Vamp::Plugin::Feature,std::allocator<Vamp::Plugin::Feature> > > >  ifeature = results->begin();
	int index = ifeature->first;
	std::vector<Vamp::Plugin::Feature,std::allocator<Vamp::Plugin::Feature> > v = ifeature->second;
	dbg("size=%i", v.size());
	if(!v.size()) return;

	__gnu_cxx::__normal_iterator<Vamp::Plugin::Feature*, std::vector<Vamp::Plugin::Feature, std::allocator<Vamp::Plugin::Feature> > > a = v.begin();
	Vamp::Plugin::Feature f = (*a);
	cout << "hasTimestamp=" << f.hasTimestamp << endl;
	*/

	//for(std::vector<Vamp::Plugin::Feature>::iterator i = feature.begin(); i!=feature.end(); i++){
	//	cout << "hasTimestamp=" << (*i).hasTimestamp << endl;
	//}

	/*
    for (unsigned int i = 0; i < results[output].size(); ++i) {
        Vamp::RealTime rt;// = Vamp::RealTime::frame2RealTime(frame, sr);
        if (results[output][i].hasTimestamp) {
            rt = results[output][i].timestamp;
        }
        cout << rt.toString() << ":";
        for (unsigned int j = 0; j < results[output][i].values.size(); ++j) {
            cout << " " << results[output][i].values[j];
        }
        cout << endl;
    }
	*/
}


void
GLScene::redraw()
{
	on_expose_event(NULL);
}


double
GLScene::get_frames_per_px()
{
	double tot_width = get_width(); //this will change when zoom is added.
	return wave->frames / tot_width;
}


void
GLScene::print(double x, double y, const char* fmt, ...)
{
	char text[256];                                              // Holds Our String
	va_list ap;                                                  // Pointer To List Of Arguments

	if (fmt == NULL)                                             // If There's No Text
		*text=0;                                                 // Do Nothing
	else {
		va_start(ap, fmt);                                       // Parses The String For Variables
		vsprintf(text, fmt, ap);                                 // And Converts Symbols To Actual Numbers
		va_end(ap);                                              // Results Are Stored In Text
	}

	glRasterPos2f(x, y);

	for (int j = 0; j < strlen(text); ++j) glCallList(m_FontListBase + text[j]);
}


double
GLScene::pos2px(gpos* pos)
{
  //converts a song position to a screen position.
  Ayyi_client::gpos_is_valid(pos);

  float ticks_per_beat_f = (float)Ayyi_client::TICKS_PER_BEAT;

  double beats_float = pos->beat + (double)pos->sub / 4.0 + ((double)pos->tick / ticks_per_beat_f);
  //printf("pos2px(): pos->tick=%.1f TICKS_PER_BEAT=%.1f\n", (double)pos->tick, (double)TICKS_PER_BEAT);

  double x = beats_float * Ayyi_client::PX_PER_BEAT;
  //char bbst[64]; pos2bbst(pos, bbst); printf("pos2px(): pos=%s beatf=%.1f x=%.1f.\n", bbst, beats_float, x);

  return x;
}


double
GLScene::px2beat(double px)
{
  //converts a canvas position to a time postition in bars.
  //the beat returned is the one that the pixel is 'in', _not_ the nearest beat boundary.

  return px / Ayyi_client::PX_PER_BEAT;
}


AppWindow::AppWindow()
  : m_VBox(false, 0), button_cancel("Cancel"), button_ok("Ok"), table(2, 2)
{
	// Top-level window.

	//set_title("Ayyi Beat Detector");
	set_title(PACKAGE_NAME);

	// Get automatically redrawn if any of their children changed allocation.
	set_reallocate_redraws(true);

	add(m_VBox);

	// Simple OpenGL scene using bitmap font.
	glScene.set_size_request(640, 240);
	table.attach(glScene, 1, 2, 0, 1);

	m_VBox.pack_start(table);
	table.attach(hbox, 1, 2, 1, 2, Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK);

	Gtk::Adjustment* adj = threshold_slider.get_adjustment();
	adj->set_value(0.5f);
	adj->set_lower(0.0f);
	adj->set_upper(1.0f);
	adj->set_step_increment(0.01f);
	adj->set_page_increment(0.1f);
	adj->changed();
	threshold_slider.set_inverted(true);

	/*
	onset_type_selector.prepend_text("hello");
	onset_type_selector.set_active_text(0);
	*/

	button_cancel.signal_clicked().connect(sigc::mem_fun(*this, &AppWindow::on_button_cancel_clicked));
	button_ok.signal_clicked().connect(sigc::mem_fun(*this, &AppWindow::on_button_ok_clicked));
	threshold_slider.signal_value_changed().connect(sigc::mem_fun(*this, &AppWindow::on_threshold_slider_changed));
	//glScene.signal_button_press_event().connect(sigc::mem_fun(glScene, &GLScene::on_button_press_event), false);

	table.attach(threshold_slider, 0, 1, 0, 1, Gtk::SHRINK);
	hbox.pack_start(zoom_slider);
	hbox.pack_start(onset_type_selector, Gtk::PACK_SHRINK);
	hbox.pack_start(progress_bar);
	hbox.pack_start(button_cancel, Gtk::PACK_SHRINK, 0);
	hbox.pack_start(button_ok, Gtk::PACK_SHRINK, 0);

	show_all();
	progress_bar.hide();
}


AppWindow::~AppWindow()
{
}


void
AppWindow::print_map()
{
	dbg(0, "map size=%i", result_cache.size());
	std::map<int, std::list<Vamp::RealTime> >::iterator i = result_cache.begin();
	for(; i != result_cache.end(); i++){
		std::list<Vamp::RealTime> list = i->second;
		printf("  key=%i size=%i : ", i->first, list.size());
		int n = 0;
		for(std::list<Vamp::RealTime>::iterator j=list.begin(); j!=list.end(); j++){
			if(n++ > 10) break;
			Vamp::RealTime rt = *j;
			cout << rt.toString() << " ";
		}
		cout << endl;
	}
}


void
AppWindow::on_aubio_init()
{
	if(!aubio){ err("aubio not initialised\n"); return; }

	//dbg(0, "getting type list...");
	std::map<std::string, int>* types = aubio->get_onset_types();
	dbg(0, "listsize=%i", types->size());
	for(std::map<std::string, int>::iterator i = types->begin(); i != types->end(); i++){
		onset_type_selector.prepend_text(i->first);
	}
	//onset_type_selector.set_active_text("Text");
	onset_type_selector.set_active(0);
}


void
AppWindow::on_button_cancel_clicked()
{
	Gtk::Main::quit();
}


void
AppWindow::on_button_ok_clicked()
{
	//make a list of new regions:
	unsigned prev = 0;
	std::list<song_object*> new_regions;
	if(result_cache.size()){
		std::list<Vamp::RealTime> list = current_lineset->second;
		int n = 1;
		std::list<Vamp::RealTime>::iterator i=list.begin();
		i++; //ignore the first item.
		for(; i!=list.end(); i++){
			Vamp::RealTime rt = *i;
			unsigned samples = rt.realTime2Frame(rt, wave->samplerate);
			dbg(0, "samples: %u --> %u", prev, samples);

			char num[8]; sprintf(num, "%02i", n);

			//make a region for this line.

			song_object* song_obj = new song_object("split" + string(num));
			song_obj->obj_type    = AYYI_OBJECT_PART;
			song_obj->from_source = false;
			song_obj->src_idx     = region->get_idx();
			//song_obj->parent_idx  = ayyi->region_get_track_num(region_idx);
			song_obj->parent_idx  = region->get_track_num();
			song_obj->len         = ayyi->samples2mu(samples - prev);
			song_obj->inset       = prev;
			ayyi->samples2pos(&song_obj->stime, prev);

			if(!song_obj->len){ err("zero length segment (%i). Ignoring...\n", n); delete song_obj; continue; }

			new_regions.push_back(song_obj);
			prev = samples;
			n++;
		}
	}

	//update the server with the new split region:
	ayyi->replace_region_with_multiple(region, new_regions);
}


void
AppWindow::on_threshold_slider_changed()
{
	static int old = 0;
	double value = threshold_slider.get_adjustment()->get_value();
	value = threshold_slider.get_value();
	int i_val = (int)(value * 100);
	//dbg("q=%p val=%p", msg_queue, GINT_TO_POINTER(i_val));
	if((i_val) && (i_val != old)){

		//check we dont already have results for this threshold value:
		Resultset_iter existing = result_cache.find(i_val);
		if(existing == window->result_cache.end()){
			//we dont have these results. Request them from the plugin:
			g_async_queue_push(msg_queue, GINT_TO_POINTER(i_val));
		}
		else{
			dbg(0, "key=%i already have that one. Using cached data...", i_val);
			current_lineset = existing;
			glScene.redraw();
		}

	}

	old = i_val;
}

